#!/bin/sh
# the next line restarts using wish \
exec wish "$0" ${1+"$@"}

#==============================================================================
# Demonstrates how to use a tablelist widget for displaying the contents of a
# directory.
#
# Copyright (c) 2010-2012  Csaba Nemethi (E-mail: csaba.nemethi@t-online.de)
#==============================================================================

package require Tk 8.3
package require tablelist 5.7
package require mentry 3.6

#
# Add some entries to the Tk option database
#
set dir [file dirname [info script]]
source [file join $dir option.tcl]

#
# Create three images
#
image create photo clsdFolderImg -file [file join $dir clsdFolder.gif]
image create photo openFolderImg -file [file join $dir openFolder.gif]
image create photo fileImg       -file [file join $dir file.gif]

#------------------------------------------------------------------------------
# displayContents
#
# Displays the contents of the directory dir in a tablelist widget.
#------------------------------------------------------------------------------
proc displayContents dir {
    #
    # Create a scrolled tablelist widget with 3 dynamic-
    # width columns and interactive sort capability
    #
    set tf .tf
    frame $tf -class ScrollArea
    set tbl $tf.tbl
    set vsb $tf.vsb
    set hsb $tf.hsb
    tablelist::tablelist $tbl \
	-columns {0 "Name"	    left
		  0 "Size"	    right
		  0 "Date Modified" left} \
	-expandcommand expandCmd -collapsecommand collapseCmd \
	-xscrollcommand [list $hsb set] -yscrollcommand [list $vsb set] \
	-movablecolumns no -showseparators yes -height 18 -width 80 \
	-populatecommand populateCmd -acceptchildcommand acceptChildCmd \
	-movablerows true -selectmode single
    if {[$tbl cget -selectborderwidth] == 0} {
	$tbl configure -spacing 1
    }
    $tbl columnconfigure 0 -formatcommand formatString -sortmode dictionary
    $tbl columnconfigure 1 -formatcommand formatSize -sortmode integer
    $tbl columnconfigure 2 -formatcommand formatString
    scrollbar $vsb -orient vertical   -command [list $tbl yview]
    scrollbar $hsb -orient horizontal -command [list $tbl xview]

    #
    # Create a pop-up menu with 2 command entries; bind the script
    # associated with its first entry to the <Double-1> event, too
    #
    set menu .menu
    menu $menu -tearoff no
    foreach {label cmd} {"Display Contents" putContentsOfSelFolder
	"Search for Pattern..." openSearchDlgForSelFolder} {
	$menu add command -label $label -command [list $cmd $tbl]
    }
    set bodyTag [$tbl bodytag]
    bind $bodyTag <<Button3>>  [bind TablelistBody <Button-1>]
    bind $bodyTag <<Button3>> +[bind TablelistBody <ButtonRelease-1>]
    bind $bodyTag <<Button3>> +[list postPopupMenu %X %Y]
    bind $bodyTag <Double-1>   [list putContentsOfSelFolder $tbl]

    #
    # Create three buttons within a frame child of the main widget
    #
    set bf .bf
    frame $bf
    set b1 $bf.b1
    set b2 $bf.b2
    set b3 $bf.b3
    button $b1 -width 10 -text "Refresh"
    button $b2 -width 10 -text "Parent"
    button $b3 -width 10 -text "Close" -command exit

    #
    # Create a pop-up menu demonstrating the supported tree styles
    #
    set menu2 .menu2
    menu $menu2 -tearoff no
    foreach {treeStyle selectBg selectFg} {
	adwaita		#4a90d9	white
	ambiance	#c8bba9	black
	aqua		#3875d7	white
	baghira		#396db5 white
	dust		#79675e	white
	dustSand	#be9a79	white
	gtk		#4a6984	white
	klearlooks	#3b8dd4	white
	mint		#9ec17a	white
	newWave		#fdbc9f	black
	oxygen1		#3b8dd4	white
	oxygen2		#3b8dd4	white
	phase		#678db2 white
	plastik		#657a9e white
	plastique	#3b8dd4 white
	radiance	#d8d7cd	black
	ubuntu		#ef7745	white
	vistaAero	#dbeffb	black
	vistaClassic	#11236a	white
	win7Aero	#dbeffb	black
	win7Classic	#11236a	white
	winnative	#000080	white
	winxpBlue	#316ac5	white
	winxpOlive	#93a070	white
	winxpSilver	#b2b4bf	black} {
	$menu2 add command -label "$tbl configure -treestyle $treeStyle ..." \
	    -command [list $tbl configure -treestyle $treeStyle \
		      -selectbackground $selectBg -selectforeground $selectFg]
    }
    bind .bf <<Button3>> [list tk_popup $menu2 %X %Y]

    #
    # Manage the widgets
    #
    grid $tbl -row 0 -rowspan 2 -column 0 -sticky news
    global winSys
    if {[string compare $winSys "aqua"] == 0} {
	grid [$tbl cornerpath] -row 0 -column 1 -sticky ew
	grid $vsb	       -row 1 -column 1 -sticky ns
    } else {
	grid $vsb -row 0 -rowspan 2 -column 1 -sticky ns
    }
    grid $hsb -row 2 -column 0 -sticky ew
    grid rowconfigure    $tf 1 -weight 1
    grid columnconfigure $tf 0 -weight 1
    pack $b1 $b2 $b3 -side left -expand yes -pady 10
    pack $bf -side bottom -fill x
    pack $tf -side top -expand yes -fill both

    #
    # Populate the tablelist with the contents of the given directory
    #
    $tbl sortbycolumn 0
    putContents $dir $tbl root
}

#------------------------------------------------------------------------------
# putContents
#
# Outputs the contents of the directory dir into the tablelist widget tbl, as
# child items of the one identified by nodeIdx.
#------------------------------------------------------------------------------
proc putContents {dir tbl nodeIdx} {
    #
    # The following check is necessary because this procedure
    # is also invoked by the "Refresh" and "Parent" buttons
    #
    if {[string compare $dir ""] != 0 &&
	(![file isdirectory $dir] || ![file readable $dir])} {
	bell
	if {[string compare $nodeIdx "root"] == 0} {
	    set choice [tk_messageBox -title "Error" -icon warning -message \
			"Cannot read directory \"[file nativename $dir]\"\
			-- replacing it with nearest existent ancestor" \
			-type okcancel -default ok]
	    if {[string compare $choice "ok"] == 0} {
		while {![file isdirectory $dir] || ![file readable $dir]} {
		    set dir [file dirname $dir]
		}
	    } else {
		return ""
	    }
	} else {
	    return ""
	}
    }

    if {[string compare $nodeIdx "root"] == 0} {
	if {[string compare $dir ""] == 0} {
	    if {[llength [file volumes]] == 1} {
		wm title . "Contents of the File System"
	    } else {
		wm title . "Contents of all File Systems"
	    }
	} else {
	    wm title . "Contents of the Directory \"[file nativename $dir]\""
	}

	$tbl delete 0 end
	set row 0
    } else {
	set row [expr {$nodeIdx + 1}]
    }

    #
    # Build a list from the data of the subdirectories and
    # files of the directory dir.  Prepend a "D" or "F" to
    # each entry's name and modification date & time, for
    # sorting purposes (it will be removed by formatString).
    #
    set itemList {}
    if {[string compare $dir ""] == 0} {
	foreach volume [file volumes] {
	    lappend itemList [list D[file nativename $volume] -1 D $volume]
	}
    } else {
	foreach entry [glob -nocomplain -types {d f} -directory $dir *] {
	    if {[catch {file mtime $entry} modTime] != 0} {
		continue
	    }

	    if {[file isdirectory $entry]} {
		lappend itemList [list D[file tail $entry] -1 \
		    D[clock format $modTime -format "%Y-%m-%d %H:%M"] $entry]
	    } else {
		lappend itemList [list F[file tail $entry] [file size $entry] \
		    F[clock format $modTime -format "%Y-%m-%d %H:%M"] ""]
	    }
	}
    }

    #
    # Sort the above list and insert it into the tablelist widget
    # tbl as list of children of the row identified by nodeIdx
    #
    set itemList [$tbl applysorting $itemList]
    $tbl insertchildlist $nodeIdx end $itemList

    #
    # Insert an image into the first cell of each newly inserted row
    #
    foreach item $itemList {
	set name [lindex $item end]
	if {[string compare $name ""] == 0} {			;# file
	    $tbl cellconfigure $row,0 -image fileImg
	} else {						;# directory
	    $tbl cellconfigure $row,0 -image clsdFolderImg
	    $tbl rowattrib $row pathName $name

	    #
	    # Mark the row as collapsed if the directory is non-empty
	    #
	    if {[file readable $name] && [llength \
		[glob -nocomplain -types {d f} -directory $name *]] != 0} {
		$tbl collapse $row
	    }
	}

	incr row
    }

    if {[string compare $nodeIdx "root"] == 0} {
	#
	# Configure the "Refresh" and "Parent" buttons
	#
	.bf.b1 configure -command [list refreshView $dir $tbl]
	set b2 .bf.b2
	if {[string compare $dir ""] == 0} {
	    $b2 configure -state disabled
	} else {
	    $b2 configure -state normal
	    set p [file dirname $dir]
	    if {[string compare $p $dir] == 0} {
		$b2 configure -command [list putContents "" $tbl root]
	    } else {
		$b2 configure -command [list putContents $p $tbl root]
	    }
	}
    }
}

#------------------------------------------------------------------------------
# formatString
#
# Returns the substring obtained from the specified value by removing its first
# character.
#------------------------------------------------------------------------------
proc formatString val {
    return [string range $val 1 end]
}

#------------------------------------------------------------------------------
# formatSize
#
# Returns an empty string if the specified value is negative and the value
# itself in user-friendly format otherwise.
#------------------------------------------------------------------------------
proc formatSize val {
    if {$val < 0} {
	return ""
    } elseif {$val < 1024} {
	return "$val bytes"
    } elseif {$val < 1048576} {
	return [format "%.1f KB" [expr {$val / 1024.0}]]
    } elseif {$val < 1073741824} {
	return [format "%.1f MB" [expr {$val / 1048576.0}]]
    } else {
	return [format "%.1f GB" [expr {$val / 1073741824.0}]]
    }
}

#------------------------------------------------------------------------------
# expandCmd
#
# Outputs the contents of the directory whose leaf name is displayed in the
# first cell of the specified row of the tablelist widget tbl, as child items
# of the one identified by row, and updates the image displayed in that cell.
#------------------------------------------------------------------------------
proc expandCmd {tbl row} {
    if {[$tbl childcount $row] == 0} {
	set dir [$tbl rowattrib $row pathName]
	putContents $dir $tbl $row
    }

    if {[$tbl childcount $row] != 0} {
	$tbl cellconfigure $row,0 -image openFolderImg
    }
}

#------------------------------------------------------------------------------
# collapseCmd
#
# Updates the image displayed in the first cell of the specified row of the
# tablelist widget tbl.
#------------------------------------------------------------------------------
proc collapseCmd {tbl row} {
    $tbl cellconfigure $row,0 -image clsdFolderImg
}

#------------------------------------------------------------------------------
# populateCmd
#
# Outputs the contents of the directory whose leaf name is displayed in the
# first cell of the specified row of the tablelist widget tbl, as child items
# of the one identified by row.
#------------------------------------------------------------------------------
proc populateCmd {tbl row} {
    set dir [$tbl rowattrib $row pathName]
    putContents $dir $tbl $row
}

#------------------------------------------------------------------------------
# acceptChildCmd
#
# Returns 1 if the specified parent node of the tablelist widget tbl accepts
# the item given by the last argument as child and 0 otherwise.
#------------------------------------------------------------------------------
proc acceptChildCmd {tbl targetParentNodeIdx sourceRow} {
    if {[string compare $targetParentNodeIdx "root"] == 0} {
	#
	# Allow only volumes as top-level items
	#
	return [expr {[$tbl depth $sourceRow] == 1}]
    } else {
	#
	# Allow only directories as parent items
	#
	return [$tbl hasrowattrib $targetParentNodeIdx pathName]
    }
}

#------------------------------------------------------------------------------
# putContentsOfSelFolder
#
# Outputs the contents of the selected folder into the tablelist widget tbl.
#------------------------------------------------------------------------------
proc putContentsOfSelFolder tbl {
    set row [$tbl curselection]
    if {[$tbl hasrowattrib $row pathName]} {		;# directory item
	set dir [$tbl rowattrib $row pathName]
	if {[file isdirectory $dir] && [file readable $dir]} {
	    if {[llength [glob -nocomplain -types {d f} -directory $dir *]]
		== 0} {
		bell
	    } else {
		putContents $dir $tbl root
	    }
	} else {
	    bell
	    tk_messageBox -title "Error" -icon error -message \
		"Cannot read directory \"[file nativename $dir]\""
	    return ""
	}
    } else {						;# file item
	bell
    }
}

#------------------------------------------------------------------------------
# openSearchDlgForSelFolder
#
# Enables the user to search the selected folder of the tablelist widget tbl.
#------------------------------------------------------------------------------
proc openSearchDlgForSelFolder tbl {
    set row [$tbl curselection]
    set key [$tbl getkeys $row]
    set top .top$key
    if {[winfo exists $top]} {
	raise $top
	return ""
    }

    toplevel $top -padx 10 -pady 10
    set dir [$tbl rowattrib $row pathName]
    wm title $top "Search in Directory \"[file nativename $dir]\""
    wm transient $top .

    #
    # "Pattern" entry
    #
    set ::data($key-pattern) ""
    set lPattern [label $top.lPattern -text "Pattern:"]
    set ePattern [entry $top.ePattern -background white \
		  -textvariable data($key-pattern)]
    set ::data($key-ePattern) $ePattern

    #
    # "Style" radiobuttons
    #
    set lStyle [label $top.lStyle -text "Style:"]
    set fStyle [frame $top.fStyle]
    set ::data($key-style) "-glob"
    set rExact  [radiobutton $fStyle.rExact  -text "Exact" \
		 -variable data($key-style) -value "-exact"]
    set rGlob   [radiobutton $fStyle.rGlob   -text "Glob" \
		 -variable data($key-style) -value "-glob"]
    set rRegexp [radiobutton $fStyle.rRegexp -text "Regexp" \
		 -variable data($key-style) -value "-regexp"]
    pack $rExact $rGlob $rRegexp -side left

    #
    # "Include subfolders" and "Ignore case" checkbuttons
    #
    set fOpts [frame $top.fOpts]
    set ::data($key-descend) 0
    set ckDescend [checkbutton $fOpts.ckDescend -text "Include subfolders" \
		   -variable data($key-descend)]
    set ::data($key-noCase) 0
    set ckNoCase [checkbutton $fOpts.ckNoCase -text "Ignore case" \
		  -variable data($key-noCase)]
    pack $ckDescend $ckNoCase -side left -padx {0 20}

    #
    # "Modification Date" checkbuttons and mentry widgets
    #
    set lfLastModDateTime [labelframe $top.lfLastModDateTime -text \
			   "Modification Date"]
    set ::data($key-useMinDateTime) 0
    set ::data($key-useMaxDateTime) 0
    set ckMinDateTime [checkbutton $lfLastModDateTime.ckMinDateTime -text \
		       "After:" -variable data($key-useMinDateTime)]
    set meMinDateTime [mentry::dateTimeMentry \
		       $lfLastModDateTime.meMinDateTime YmdHM - : \
		       -justify center -background white]
    set ckMaxDateTime [checkbutton $lfLastModDateTime.ckMaxDateTime -text \
		       "Before:" -variable data($key-useMaxDateTime)]
    set meMaxDateTime [mentry::dateTimeMentry \
		       $lfLastModDateTime.meMaxDateTime YmdHM - : \
		       -justify center -background white]
    set maxClock [clock seconds]
    set minClock [expr {$maxClock - 24*60*60}]
    mentry::putClockVal $minClock $meMinDateTime
    mentry::putClockVal $maxClock $meMaxDateTime
    $meMinDateTime configure -state disabled
    $meMaxDateTime configure -state disabled
    $ckMinDateTime configure -command \
	[list setEnabled $meMinDateTime $ckMinDateTime]
    $ckMaxDateTime configure -command \
	[list setEnabled $meMaxDateTime $ckMaxDateTime]
    grid $ckMinDateTime -row 0 -column 0 -sticky w -padx {0 3} -pady {3 6}
    grid $meMinDateTime -row 0 -column 1 -sticky w -padx {0 6} -pady {3 6}
    grid $ckMaxDateTime -row 1 -column 0 -sticky w -padx {0 3} -pady {0 6}
    grid $meMaxDateTime -row 1 -column 1 -sticky w -padx {0 6} -pady {0 6}
    set ::data($key-meMinDateTime) $meMinDateTime
    set ::data($key-meMaxDateTime) $meMaxDateTime

    #
    # "Find" and "Close" buttons
    #
    set fBtns [frame $top.fBtns]
    set bFind  [button $fBtns.bFind -text "Find" -default active \
		-command [list findPattern $tbl $key]]
    set bClose [button $fBtns.bClose -text "Close" -default normal \
		-command [list destroy $top]]
    bind $bClose <FocusIn> [format {
	%%W configure -default active
	%s configure -default normal
    } $bFind]
    bind $bClose <FocusOut> [format {
	if {[string compare [focus -lastfor %%W] %%W] != 0} {
	    %%W configure -default normal
	    %s configure -default active
	}
    } $bFind]
    pack $bFind $bClose -pady {0 2} -expand yes -fill x
    set ::data($key-bFind) $bFind
    set ::data($key-bClose) $bClose

    #
    # Readonly text widget displaying the search command
    #
    set tCmd [text $top.tCmd -background white -height 2 -width 70 -wrap word \
	      -highlightthickness 0 -insertwidth 0]
    wcb::callback $tCmd before insert cancelInput
    wcb::callback $tCmd before delete cancelInput
    proc cancelInput {w idx args} { wcb::cancel }
    set ::data($key-tCmd) $tCmd

    #
    # Scrollable tablelist widget displaying the search result
    #
    set fResult $top.fResult
    frame $fResult -class ScrollArea
    set resTbl $fResult.tbl
    set vsb $fResult.vsb
    set hsb $fResult.hsb
    tablelist::tablelist $resTbl \
	-columns {0 "Path"	    left
		  0 "Size"	    right
		  0 "Date Modified" left} \
	-xscrollcommand [list $hsb set] -yscrollcommand [list $vsb set] \
	-movablecolumns no -showseparators yes -height 12 -width 70
    if {[$resTbl cget -selectborderwidth] == 0} {
	$resTbl configure -spacing 1
    }
    $resTbl columnconfigure 0 -formatcommand formatString -sortmode dictionary
    $resTbl columnconfigure 1 -formatcommand formatSize -sortmode integer
    $resTbl columnconfigure 2 -formatcommand formatString
    scrollbar $vsb -orient vertical   -command [list $resTbl yview]
    scrollbar $hsb -orient horizontal -command [list $resTbl xview]
    grid $resTbl -row 0 -rowspan 2 -column 0 -sticky news
    global winSys
    if {[string compare $winSys "aqua"] == 0} {
	grid [$resTbl cornerpath] -row 0 -column 1 -sticky ew
	grid $vsb		  -row 1 -column 1 -sticky ns
    } else {
	grid $vsb -row 0 -rowspan 2 -column 1 -sticky ns
    }
    grid $hsb -row 2 -column 0 -sticky ew
    grid rowconfigure    $fResult 1 -weight 1
    grid columnconfigure $fResult 0 -weight 1
    set ::data($key-resTbl) $resTbl

    #
    # Create bindings for <Return>, <Escape>, and <Destroy>
    #
    bind $top <Return> [list invokeActiveBtn %W]
    bind $top <Escape> [list destroy $top]
    bind $top <Destroy> [list unsetData %W]

    #
    # Manage the widgets
    #
    grid $lPattern -row 0 -column 0 -sticky w -padx {0 3} -pady {0 6}
    grid $ePattern -row 0 -column 1 -sticky ew -padx {0 10} -pady {0 6}
    grid $lStyle -row 1 -column 0 -sticky w -padx {0 3} -pady {0 6}
    grid $fStyle -row 1 -column 1 -sticky w -padx {0 10} -pady {0 6}
    grid $fOpts -row 2 -column 0 -columnspan 2 -sticky w -pady {0 6}
    grid $lfLastModDateTime -row 0 -column 2 -rowspan 3 -sticky w \
	 -padx {0 10} -pady {0 6}
    grid $tCmd -row 3 -column 0 -columnspan 5 -sticky ew -pady {0 6}
    grid $fBtns -row 0 -column 4 -rowspan 3 -sticky ne
    grid $fResult -row 4 -column 0 -columnspan 5 -sticky news
    grid rowconfigure    $top 4 -weight 1
    grid columnconfigure $top 3 -weight 1

    focus $ePattern
}

#------------------------------------------------------------------------------
# setEnabled
#
# Enables or disables the widget w, depending on the checked state of the
# checkbutton ckBtn.
#------------------------------------------------------------------------------
proc setEnabled {w ckBtn} {
    set varName [$ckBtn cget -variable]
    if {[set ::$varName]} {
	$w configure -state normal
    } else {
	$w configure -state disabled
    }
}

#------------------------------------------------------------------------------
# findPattern
#
# Searches for the pattern entered into the search dialog corresponding to the
# specified key of the tablelist widget tbl.
#------------------------------------------------------------------------------
proc findPattern {tbl key} {
    #
    # Clear the readonly text widget displaying the search command
    #
    set tCmd $::data($key-tCmd)
    _$tCmd delete 1.0 end

    #
    # Clear the tablelist widget displaying the search result
    #
    set resTbl $::data($key-resTbl)
    $resTbl delete 0 end
    $resTbl resetsortinfo

    #
    # Check the pattern
    #
    if {[string compare $::data($key-style) "-regexp"] == 0 &&
	[catch {regexp $::data($key-pattern) ""}] != 0} {
	bell
	tk_messageBox -parent .top$key -title "Error" -icon error \
		      -message "Invalid regular expression pattern"
	focus $::data($key-ePattern)
	return ""
    }

    #
    # Message strings corresponding to the values
    # returned by mentry::getClockVal on failure
    #
    array set msgs {
	EMPTY		"Field value missing"
	BAD		"Invalid field value"
	BAD_DATE	"Invalid date"
	BAD_YEAR	"Unsupported year"
    }

    set useMinDateTime $::data($key-useMinDateTime)
    set minDateTime ""
    if {$useMinDateTime} {
	if {[catch {
	    set clockVal [mentry::getClockVal $::data($key-meMinDateTime)]
	} result] == 0} {
	    set minDateTime [clock format $clockVal -format "%Y-%m-%d %H:%M"]
	} else {
	    bell
	    tk_messageBox -parent .top$key -title "Error" -icon error \
			  -message $msgs($result)
	    return ""
	}
    }

    set useMaxDateTime $::data($key-useMaxDateTime)
    set maxDateTime ""
    if {$useMaxDateTime} {
	if {[catch {
	    set clockVal [mentry::getClockVal $::data($key-meMaxDateTime)]
	} result] == 0} {
	    set maxDateTime [clock format $clockVal -format "%Y-%m-%d %H:%M"]
	} else {
	    bell
	    tk_messageBox -parent .top$key -title "Error" -icon error \
			  -message $msgs($result)
	    return ""
	}
    }

    #
    # Build, show, and evaluate the search command
    #
    set searchCmd [list $tbl searchcolumn 0 $::data($key-pattern) \
		   -parent k$key -all -formatted $::data($key-style)]
    if {$::data($key-descend)} {
	lappend searchCmd -descend
    }
    if {$::data($key-noCase)} {
	lappend searchCmd -nocase
    }
    if {$useMinDateTime || $useMaxDateTime} {
	set checkCmd [list checkDateTime $minDateTime $maxDateTime]
	lappend searchCmd -check $checkCmd
    }
    _$tCmd insert end $searchCmd
    update idletasks
    set rowList [eval $searchCmd]

    #
    # Populate the search result tablelist with the data of the matching items
    #
    foreach row $rowList {
	set item [$tbl get $row]
	foreach {name size dateTime} $item {}

	if {[$tbl hasrowattrib $row pathName]} {	;# directory item
	    set prefix D
	    set img clsdFolderImg
	} else {					;# file item
	    set prefix F
	    set img fileImg
	}

	set relPath [string range $name 1 end]
	set parentKey [$tbl parentkey $row]
	while {[string compare $parentKey k$key] != 0} {
	    set parentName [$tbl getformattedcells $parentKey,0]
	    set relPath [file join $parentName $relPath]
	    set parentKey [$tbl parentkey $parentKey]
	}
	set relPath $prefix[file nativename $relPath]

	$resTbl insert end [list $relPath $size $dateTime]
	$resTbl cellconfigure end,0 -image $img
    }
}

#------------------------------------------------------------------------------
# checkDateTime
#
# Checks the last modification date & time of the directory entry displayed in
# the specified row of the tablelist widget tbl.
#------------------------------------------------------------------------------
proc checkDateTime {minDateTime maxDateTime tbl row col elem} {
    set dateTime [$tbl getformattedcells $row,2]
    if {$minDateTime ne "" && [string compare $dateTime $minDateTime] < 0} {
	return 0
    }
    if {$maxDateTime ne "" && [string compare $dateTime $maxDateTime] > 0} {
	return 0
    }

    return 1
}

#------------------------------------------------------------------------------
# invokeActiveBtn
#
# Invokes the current active button of the toplevel window containing the
# specified widget.
#------------------------------------------------------------------------------
proc invokeActiveBtn w {
    if {[regexp {^(Checkbutton|Radiobutton)$} [winfo class $w]]} {
	return ""
    }

    set top [winfo toplevel $w]
    set key [string range $top 4 end]
    foreach btn [list $::data($key-bFind) $::data($key-bClose)] {
	if {[string compare [$btn cget -default] "active"] == 0} {
	    $btn invoke
	    break
	}
    }
}

#------------------------------------------------------------------------------
# unsetData
#
# Unsets the elements of the array ::data whose names are associated with the
# specified widget.
#------------------------------------------------------------------------------
proc unsetData w {
    if {[string compare $w [winfo toplevel $w]] != 0} {
	return ""
    }

    set key [string range $w 4 end]
    foreach name [array names ::data $key-*] {
	unset ::data($name)
    }
}

#------------------------------------------------------------------------------
# postPopupMenu
#
# Posts the pop-up menu .menu at the given screen position.  Before posting
# the menu, the procedure enables/disables its only entry, depending upon
# whether the selected item represents a readable directory or not.
#------------------------------------------------------------------------------
proc postPopupMenu {rootX rootY} {
    set tbl .tf.tbl
    set row [$tbl curselection]
    set menu .menu
    if {[$tbl hasrowattrib $row pathName]} {		;# directory item
	set dir [$tbl rowattrib $row pathName]
	if {[file isdirectory $dir] && [file readable $dir]} {
	    if {[llength [glob -nocomplain -types {d f} -directory $dir *]]
		== 0} {
		foreach idx {0 1 2 3} {
		    $menu entryconfigure $idx -state disabled
		}
	    } else {
		foreach idx {0 1 2 3} {
		    $menu entryconfigure $idx -state normal
		}
	    }
	} else {
	    bell
	    tk_messageBox -title "Error" -icon error -message \
		"Cannot read directory \"[file nativename $dir]\""
	    return ""
	}
    } else {						;# file item
	foreach idx {0 1 2 3} {
	    $menu entryconfigure $idx -state disabled
	}
    }

    tk_popup $menu $rootX $rootY
}

#------------------------------------------------------------------------------
# refreshView
#
# Redisplays the contents of the directory dir in the tablelist widget tbl and
# restores the expanded states of the folders as well as the vertical view.
#------------------------------------------------------------------------------
proc refreshView {dir tbl} {
    #
    # Save the vertical view and get the path names
    # of the folders displayed in the expanded rows
    #
    set yView [$tbl yview]
    foreach key [$tbl expandedkeys] {
	set pathName [$tbl rowattrib $key pathName]
	set expandedFolders($pathName) 1
    }

    #
    # Redisplay the directory's (possibly changed) contents and restore
    # the expanded states of the folders, along with the vertical view
    #
    putContents $dir $tbl root
    restoreExpandedStates $tbl root expandedFolders
    $tbl yview moveto [lindex $yView 0]
}

#------------------------------------------------------------------------------
# restoreExpandedStates
#
# Expands those children of the parent identified by nodeIdx that display
# folders whose path names are the names of the elements of the array specified
# by the last argument.
#------------------------------------------------------------------------------
proc restoreExpandedStates {tbl nodeIdx expandedFoldersName} {
    upvar $expandedFoldersName expandedFolders

    foreach key [$tbl childkeys $nodeIdx] {
	set pathName [$tbl rowattrib $key pathName]
	if {[string compare $pathName ""] != 0 &&
	    [info exists expandedFolders($pathName)]} {
	    $tbl expand $key -partly
	    restoreExpandedStates $tbl $key expandedFolders
	}
    }
}

displayContents ""
